#include "c_battle.h"
#include "c_map.h"
#include "c_options.h"
#include <windows.h>

static void map_reset(map_t);
static BOOL map_cell_is_near(map_t, POINT, cell_t);
static BOOL map_fit_ship(map_t, POINT, int, dir_t);

static void map_reset(map_t map) {
	int x, y;

	for (y = 0; y < NUM_CELLS; y++)
		for (x = 0; x < NUM_CELLS; x++)
			map[y][x] = ctFree;
}

static BOOL map_cell_is_near(map_t map, POINT cell, cell_t who) {
	POINT dummy;

	return !!map_cell_iterator(map, cell, &dummy, 0, who);
}

static BOOL map_fit_ship(map_t map, POINT cell, int ship_kind, dir_t dir) {
	int i;
	POINT pt;
	BOOL retval = TRUE;

	if (map[cell.y][cell.x] == ctShip)
		retval = FALSE;
	else {
		switch (dir) {
			case dirHorz:
				if (cell.x + ship_kind > NUM_CELLS)
					retval = FALSE;
				else {
					for (i = cell.x; i < cell.x + ship_kind; i++) {
						pt.x = i, pt.y = cell.y;
						if (map_cell_is_near(map, pt, ctShip))
							retval = FALSE;
					}
				}
				if (retval)
					for (i = cell.x; i < cell.x + ship_kind; i++)
						map[cell.y][i] = ctShip;
				break;
			case dirVert:
				if (cell.y + ship_kind > NUM_CELLS)
					retval = FALSE;
				else {
					for (i = cell.y; i < cell.y + ship_kind; i++) {
						pt.x = cell.x, pt.y = i;
						if (map_cell_is_near(map, pt, ctShip))
							retval = FALSE;
					}
				}
				if (retval)
					for (i = cell.y; i < cell.y + ship_kind; i++)
						map[i][cell.x] = ctShip;
				break;
		}
	}

	return retval;
}

void map_randomize(map_t map) {
	int ship_kind, ship_num;
	POINT cell;

	map_reset(map);
	for (ship_kind = NUM_SHIPS; ship_kind >= 1; ship_kind--) {
		for (ship_num = 0; ship_num < NUM_SHIPS - ship_kind + 1; ship_num++) {
			do {
				cell.x = rand() % NUM_CELLS;
				cell.y = rand() % NUM_CELLS;
			} while (!map_fit_ship(map, cell, ship_kind, rand() % NUM_DIRS));
		}
	}
}

void map_refresh(map_t map, int cell_size, HDC hdc, int map_type, const COLORREF color[NUM_COLORS]) {
	int x, y;

	for (y = 0; y < NUM_CELLS; y++)
		for (x = 0; x < NUM_CELLS; x++) {
			RECT r;
			HBRUSH hbrush;

			r.left = (map_type == ENEMY ? x : x + NUM_CELLS + 1) * cell_size + 2;
			r.top = y * cell_size + 2;
			r.right = (map_type == ENEMY ? x : x + NUM_CELLS + 1) * cell_size + cell_size - 1;
			r.bottom = y * cell_size + cell_size - 1;
			switch (map[y][x]) {
				case ctShip:
				case ctHit:
					if (map_type != ENEMY || map[y][x] == ctHit) {
						/*       ...  ? :) */
						hbrush = CreateSolidBrush(color[clrShip]);
						FillRect(hdc, &r, hbrush);
						DeleteObject(hbrush);
					}
					if (map[y][x] == ctHit) {
						HPEN holdpen = SelectObject(hdc,
								CreatePen(PS_SOLID, 1, color[clrHit]));
						MoveToEx(hdc, r.left, r.top, NULL);
						LineTo(hdc, r.right, r.bottom);
						MoveToEx(hdc, r.left, r.bottom, NULL);
						LineTo(hdc, r.right, r.top);
						DeleteObject(SelectObject(hdc, holdpen));
					}
					break;
				case ctMiss:
					hbrush = SelectObject(hdc, CreateSolidBrush(color[clrMiss]));
					InflateRect(&r, -((r.right - r.left) / 3), -((r.bottom - r.top) / 3));
					Ellipse(hdc, r.left, r.top, r.right, r.bottom);
					DeleteObject(SelectObject(hdc, hbrush));
					break;
			}
		}
}

int map_sink_ship(map_t map, POINT cell) {
	POINT beg, end;
	BOOL horizontal;
	int retval;

	ASSERT(map[cell.y][cell.x] == ctHit);

	/* simple case */
	if (map_cell_is_near(map, cell, ctShip))
		return 0;

	beg = end = cell;
	map_cell_iterator(map, beg, &end, 0, ctHit);
	horizontal = beg.x != end.x;
	if (horizontal) {
		while (beg.x > 0 && map[beg.y][--beg.x] == ctHit) ;
		while (end.x < NUM_CELLS - 1 && map[end.y][++end.x] == ctHit) ;
		if (map[beg.y][beg.x] == ctShip || map[end.y][end.x] == ctShip)
			return FALSE;
		if (map[beg.y][beg.x] != ctHit)
			beg.x++;
		if (map[end.y][end.x] != ctHit)
			end.x--;
		retval = beg.x;
		while (beg.x <= end.x) {
			int i = 0;
			POINT pt;
			while ((i = map_cell_iterator(map, beg, &pt, i, ctFree)) != 0)
				map[pt.y][pt.x] = ctMiss;
			beg.x++;
		}
		retval = end.x - retval + 1;
	} else {
		while (beg.y > 0 && map[--beg.y][beg.x] == ctHit) ;
		while (end.y < NUM_CELLS - 1 && map[++end.y][end.x] == ctHit) ;
		if (map[beg.y][beg.x] == ctShip || map[end.y][end.x] == ctShip)
			return FALSE;
		if (map[beg.y][beg.x] != ctHit)
			beg.y++;
		if (map[end.y][end.x] != ctHit)
			end.y--;
		retval = beg.y;
		while (beg.y <= end.y) {
			int i = 0;
			POINT pt;
			while ((i = map_cell_iterator(map, beg, &pt, i, ctFree)) != 0)
				map[pt.y][pt.x] = ctMiss;
			beg.y++;
		}
		retval = end.y - retval + 1;
	}

	return horizontal ? retval : -retval;
}

BOOL map_all_sunk(map_t map) {
	int x, y;

	for (y = 0; y < NUM_CELLS; y++)
		for (x = 0; x < NUM_CELLS; x++)
			if (map[y][x] == ctShip)
				return FALSE;
	return TRUE;
}

int map_cell_iterator(map_t map, POINT cell, POINT *next, int offset, cell_t who) {

	switch (offset) {
		case 0:
			if (cell.y > 0 && cell.x > 0 && map[cell.y - 1][cell.x - 1] == who) {
				next->y = cell.y - 1, next->x = cell.x - 1;
				return 1;
			}
		case 1:
			if (cell.y > 0 && map[cell.y - 1][cell.x] == who) {
				next->y = cell.y - 1, next->x = cell.x;
				return 2;
			}
		case 2:
			if (cell.y > 0 && cell.x < NUM_CELLS - 1 && map[cell.y - 1][cell.x + 1] == who) {
				next->y = cell.y - 1, next->x = cell.x + 1;
				return 3;
			}
		case 3:
			if (cell.x < NUM_CELLS - 1 && map[cell.y][cell.x + 1] == who) {
				next->y = cell.y, next->x = cell.x + 1;
				return 4;
			}
		case 4:
			if (cell.y < NUM_CELLS - 1 && cell.x < NUM_CELLS - 1 && map[cell.y + 1][cell.x + 1] == who) {
				next->y = cell.y + 1, next->x = cell.x + 1;
				return 5;
			}
		case 5:
			if (cell.y < NUM_CELLS - 1 && map[cell.y + 1][cell.x] == who) {
				next->y = cell.y + 1, next->x = cell.x;
				return 6;
			}
		case 6:
			if (cell.y < NUM_CELLS - 1 && cell.x > 0 && map[cell.y + 1][cell.x - 1] == who) {
				next->y = cell.y + 1, next->x = cell.x - 1;
				return 7;
			}
		case 7:
			if (cell.x > 0 && map[cell.y][cell.x - 1] == who) {
				next->y = cell.y, next->x = cell.x - 1;
				return 8;
			}
	}
	return 0;
}
